"""
Code created by Sorin Liviu Jurj and Allen-Jasmin Farcas for the paper called "Mobile Application for Receipt Fraud Detection Based on Optical Character Recognition".
More information about the paper can be found here: https://www.jurj.de/mobile-application-for-receipt-fraud-detection-based-on-optical-character-recognition/
"""
import time
from PIL import Image
import pytesseract
import cv2
import os
import matplotlib.pyplot as plt

def threshold(img):
	img = img.copy()
	img = img.astype('uint8')
	img = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
	height , width = img.shape
	img = cv2.bitwise_not(img)
	blurSize = 2*int(width*0.01/2)+1
	threshParameter = 2*int(width*0.9/2)+1
	img = cv2.GaussianBlur(img, (blurSize, blurSize), 0)
	img = cv2.adaptiveThreshold(img, 255, cv2.ADAPTIVE_THRESH_GAUSSIAN_C, cv2.THRESH_BINARY, threshParameter, -20)
	return img

def show(img):
	plt.imshow(img,  cmap = 'gray', interpolation = 'bicubic')
	plt.xticks([]), plt.yticks([])  # to hide tick values on X and Y axis
	plt.show()

def test_ocr_digits():
	directory =  'data/cropped'
	with open("results_tesseract_scores_digits.csv", 'w') as w:
		w.write("Number;OCR_thresh;thresh_score;OCR_no_thresh;no_thresh_score\n")
		score_th = 0
		score_n_th = 0
		total = 0.0
		suma = 0.0
		nr = 0.0
		for filename in os.listdir(directory):
			total += 1
			img = cv2.imread(os.path.join(directory, filename))
			fname = filename.split(".")[0]
			fname = fname.split("_")
			fn = "{}".format(fname[0])
			start = time.time()
			text_no_thresh = pytesseract.image_to_string(Image.fromarray(img))
			stop = time.time()
			suma += stop-start
			nr += 1
			if fn != text_no_thresh:
				score_n_th += 1
				no_thresh_score = 1
			else :
				no_thresh_score = 0
			text_thresh = ""
			thresh_score = ""
			w.write("{};{};{};{};{}\n".format(fn,text_thresh,thresh_score,text_no_thresh,no_thresh_score))
		print("Average time {} sum {} number {}".format(suma/nr,suma,nr))
		w.write("{};{};{};{};{}\n".format("","WRONG",score_th,"WRONG",score_n_th))
		w.write("{};{};{};{};{}\n".format("","WRONG",(score_th/total)*100,"WRONG",(score_n_th/total)*100))
		w.write("{};{};{};{};{}\n".format("","RIGHT",100-score_th/total*100,"RIGHT",100-score_n_th/total*100))


def test_ocr_receipt():
	directory =  'receiptData/photos'
	with open("results_tesseract_scores_receipt.csv", 'w') as w:
		w.write("Number;OCR_thresh;OCR_no_thresh\n")
		suma = 0.0
		nr = 0.0
		for filename in os.listdir(directory):
			img = cv2.imread(os.path.join(directory, filename))
			fname = filename.split(".")[0]
			fname = fname.split("_")
			fn = "{}".format(fname[0])
			start = time.time()
			text_no_thresh = pytesseract.image_to_string(Image.fromarray(img))
			stop = time.time()
			suma += stop-start
			nr += 1
			text_no_thresh = text_no_thresh.split("\n")
			text_thresh = ""
			w.write("{};{};{}\n".format(fname,text_thresh,text_no_thresh))
		print("Average time {} sum {} number {}".format(suma/nr,suma,nr))

def result_ocr():
	start = time.time()
	test_ocr_digits()
	end = time.time()
	print(end - start)

	start = time.time()
	test_ocr_receipt()
	end = time.time()
	print(end - start)

result_ocr()
